import os
import sys
import shlex

from ccinput.__init__ import __version__
from ccinput.packages.gaussian import GaussianCalculation
from ccinput.packages.orca import OrcaCalculation
from ccinput.packages.xtb import XtbCalculation
from ccinput.drivers.pysis import PysisDriver

from ccinput.calculation import (
    Calculation,
    Parameters,
    Constraint,
    parse_str_constraints,
    parse_freeze_constraints,
    parse_scan_constraints,
)
from ccinput.utilities import (
    get_abs_type,
    get_abs_software,
    standardize_xyz,
    parse_xyz_from_file,
    warn,
)
from ccinput.exceptions import *
from ccinput.presets import (
    save_preset,
    load_preset,
    list_presets,
    print_preset,
    is_preset,
)

SOFTWARE_CLASSES = {
    "gaussian": GaussianCalculation,
    "orca": OrcaCalculation,
    "xtb": XtbCalculation,
    "pysis": PysisDriver,
    "Pysisyphus": PysisDriver,
}


def process_calculation(calc):
    if calc.driver in ["none", None, ""]:
        cls = SOFTWARE_CLASSES[calc.parameters.software](calc)
    else:
        cls = SOFTWARE_CLASSES[calc.driver](calc)
    return cls


def generate_calculation(
    software=None,
    type=None,
    method="",
    basis_set="",
    solvent="",
    solvation_model="",
    solvation_radii="",
    custom_solvation_radii="",
    specifications="",
    freeze=[],
    scan=[],
    sfrom=[],
    sto=[],
    snsteps=[],
    sstep=[],
    density_fitting="",
    custom_basis_sets="",
    xyz="",
    constraints="",
    nproc=1,
    mem=1000,
    charge=0,
    parse_name=False,
    multiplicity=1,
    d3=False,
    d3bj=False,
    aux_name=None,
    name=None,
    header="File created by ccinput",
    file=None,
    driver="none",
    trust_me=False,
    **kwargs,
):
    if software is None:
        raise InvalidParameter("Specify a software package to use")

    if type is None:
        raise InvalidParameter("Specify a calculation type")

    if xyz == "":
        raise InvalidParameter("No input structure")

    xyz_structure = standardize_xyz(xyz)

    abs_software = get_abs_software(software)

    if method is None:
        raise InvalidParameter("Specify a calculation method")

    calc_type = get_abs_type(type)

    params = Parameters(
        abs_software,
        solvent,
        solvation_model,
        solvation_radii,
        custom_solvation_radii,
        basis_set,
        method,
        specifications,
        density_fitting,
        custom_basis_sets,
        d3,
        d3bj,
        trust_me,
        **kwargs,
    )

    _constraints = parse_str_constraints(
        constraints, xyz_structure, software=abs_software
    )
    _constraints += parse_freeze_constraints(
        freeze, xyz_structure, software=abs_software
    )
    _constraints += parse_scan_constraints(
        scan, sfrom, sto, snsteps, sstep, xyz_structure, software=abs_software
    )

    calc = Calculation(
        xyz_structure,
        params,
        calc_type,
        constraints=_constraints,
        nproc=nproc,
        mem=mem,
        charge=charge,
        multiplicity=multiplicity,
        parse_name=parse_name,
        aux_name=aux_name,
        name=name,
        header=header,
        software=abs_software,
        file=file,
        driver=driver,
        **kwargs,
    )

    return process_calculation(calc)


def gen_obj(**args):
    if "file" in args:
        if args["file"] is None:
            pass
        elif isinstance(args["file"], list):
            if len(args["file"]) > 1:
                print("file", args["file"])
                raise UnimplementedError(
                    "No support for multiple input files at once except from the command line"
                )
            else:
                xyz = parse_xyz_from_file(args["file"][0])
                args["xyz"] = xyz
        else:
            xyz = parse_xyz_from_file(args["file"])
            args["xyz"] = xyz

    return generate_calculation(**args)


def gen_input(**args):
    return gen_obj(**args).output


def write_input(filename, **args):
    inp = gen_input(**args)
    with open(filename, "w") as out:
        out.write(inp)


def get_parser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Generates an input for a computational chemistry package"
    )
    parser.add_argument(
        "software", nargs="?", help="Desired software package (Gaussian or ORCA)"
    )

    parser.add_argument("type", nargs="?", help="Calculation type (opt, freq, sp, ...)")

    parser.add_argument(
        "method", nargs="?", help="Computational method (HF, AM1, B3LYP, ...)"
    )

    parser.add_argument("--basis_set", "-bs", default="", type=str, help="Basis set")

    parser.add_argument(
        "--solvent", "-s", default="", type=str, help="Solvent for implicit solvation"
    )

    parser.add_argument(
        "--solvation_model",
        "-sm",
        default="",
        type=str,
        help="Solvation model for implicit solvation",
    )

    parser.add_argument(
        "--solvation_radii",
        "-sr",
        default="",
        type=str,
        help="Set of solvation radii for implicit solvation",
    )

    parser.add_argument(
        "--custom_solvation_radii",
        "-csr",
        default="",
        type=str,
        help="Specific solvation radii to modify",
    )

    parser.add_argument(
        "--specifications", "-spec", default="", type=str, help="Additional commands"
    )

    parser.add_argument(
        "--density_fitting",
        "-df",
        default="",
        type=str,
        help="Basis set for density fitting (Gaussian only)",
    )

    parser.add_argument(
        "--custom_basis_sets",
        "-cbs",
        default="",
        type=str,
        help="String specification of custom basis sets on a per-atom basis",
    )

    parser.add_argument(
        "--xyz", "-x", default="", type=str, help="XYZ structure as string"
    )

    parser.add_argument(
        "--file",
        "-f",
        default=None,
        nargs="+",
        type=str,
        help="XYZ structure(s) as file(s)",
    )

    parser.add_argument(
        "--output",
        "-o",
        default="",
        type=str,
        help="Write the result to the specified file (single file) or using the specified pattern (multiple files)",
    )

    parser.add_argument(
        "--constraints",
        "-co",
        default="",
        type=str,
        help="String specification of constraints for certain calculations",
    )

    parser.add_argument(
        "--freeze",
        action="append",
        nargs="+",
        default=[],
        metavar="ATOM",
        help="Freeze specified distance/angle/dihedral between the specified atoms",
    )

    parser.add_argument(
        "--scan",
        action="append",
        nargs="+",
        default=[],
        metavar="ATOM",
        help="Scan specified distance/angle/dihedral between the specified atoms (additional parameters required)",
    )

    parser.add_argument(
        "--from",
        action="append",
        dest="sfrom",
        default=[],
        metavar="FROM",
        help="Initial distance (for --scan)",
    )

    parser.add_argument(
        "--to",
        action="append",
        dest="sto",
        default=[],
        metavar="TO",
        help="Final distance (for --scan)",
    )

    parser.add_argument(
        "--nsteps",
        action="append",
        dest="snsteps",
        default=[],
        metavar="NSTEPS",
        help="Number of steps (for --scan)",
    )

    parser.add_argument(
        "--step",
        action="append",
        dest="sstep",
        default=[],
        metavar="STEP",
        help="Step size in Å (for --scan)",
    )

    parser.add_argument(
        "--nproc", "-n", default=1, type=int, help="Number of CPU cores to use"
    )

    parser.add_argument(
        "--mem",
        default="1000MB",
        help="Amount of memory to use (if no unit is provided, assumed to be in MB)",
    )

    parser.add_argument(
        "--charge", "-c", default=0, type=int, help="Charge of the system"
    )

    parser.add_argument(
        "--mult", "-m", default=1, type=int, help="Multiplicity of the system"
    )

    parser.add_argument(
        "--parse_name",
        action="store_true",
        help="Use filenames to assign the charge and multiplicity",
    )

    parser.add_argument(
        "--trust_me",
        action="store_true",
        help="Unknown values will be accepted as given",
    )

    dispersion_group = parser.add_mutually_exclusive_group()

    dispersion_group.add_argument(
        "--d3", action="store_true", help="Use the D3(0) dispersion " "correction"
    )

    dispersion_group.add_argument(
        "--d3bj",
        action="store_true",
        help="Use the D3 dispersion " "correction with Becke-Johnson damping",
    )

    parser.add_argument(
        "--name",
        default=None,
        type=str,
        help="Name of the produced file (unused by some packages)",
    )

    parser.add_argument(
        "--aux_name",
        default=None,
        type=str,
        help="Name of the auxiliary file (some calculation types only)",
    )

    parser.add_argument(
        "--header",
        default="File created by ccinput",
        type=str,
        help="Header in produced file (unused by some packages)",
    )

    parser.add_argument(
        "--save",
        type=str,
        help="Save the current parameters as a preset of the given name",
    )

    parser.add_argument(
        "--preset",
        type=str,
        nargs="?",
        const="",
        help="Load parameters from the chosen preset",
    )
    parser.add_argument(
        "--driver",
        choices=["none", "ORCA", "pysis"],
        default="none",
        type=str.lower,
        help="Specify a computation driver other than the calculation package",
    )
    parser.add_argument(
        "--version", "-v", action="version", version=f"%(prog)s {__version__}"
    )
    return parser


def cmd(cmd_line=None):
    parser = get_parser()

    if cmd_line:
        args = parser.parse_args(shlex.split(cmd_line))
    else:
        args = parser.parse_args()

    if args.save:
        if args.preset:
            warn("Cannot both save and load a preset")
            exit(0)
        preset_name = save_preset(args, parser.parse_args(["", "", ""]))
        print_preset(preset_name)
        return

    if (args.preset and not is_preset(args.preset)) or args.preset == "":
        list_presets()
        exit(0)

    if args.preset:
        calcs, outputs = get_input_from_args(
            args, default_params=vars(parser.parse_args([]))
        )
    else:
        calcs, outputs = get_input_from_args(args)

    if args.output != "":
        for calc, outp in zip(calcs, outputs):
            with open(outp, "w") as out:
                out.write(calc.input_file)
            if hasattr(calc, "command") and calc.command:
                print(f'Input file written to {outp} - run "{calc.command}"')
            else:
                print(f"Input file written to {outp}")
    else:
        if len(calcs) == 1:
            print(calcs[0].output)
        else:
            for calc in calcs:
                n = max(int((39 - len(calc.calc.name)) / 2), 4)
                header = "-" * n + f" {calc.calc.name} " + "-" * n
                print(header)
                print(calc.output)
                print("\n\n")


def get_input_from_args(args, default_params=None):
    xyzs = []
    names = []
    outputs = []
    files = []

    if args.file:
        xyzs = [parse_xyz_from_file(f) for f in args.file]
        files = args.file
        if len(args.file) > 1 or args.name is None:
            names = [os.path.basename(f).split(".")[0] for f in args.file]
        else:
            names = [args.name]

        if args.output != "":
            head, tail = os.path.split(args.output)
            if tail.find(".") != -1:
                has_ext = True
                prefix, ext = tail.split(".")
            else:
                has_ext = False
                prefix = tail

            if prefix != "":
                prefix += "_"
            if len(args.file) > 1:
                if has_ext:
                    outputs = [
                        os.path.join(head, prefix + name + "." + ext) for name in names
                    ]
                else:
                    outputs = [os.path.join(head, prefix + name) for name in names]
            else:
                outputs = [args.output]
    else:
        xyzs = [args.xyz]
        names = [args.name]
        outputs = [args.output]
        files = [args.file]

    params = {
        "software": args.software,
        "type": args.type,
        "method": args.method,
        "basis_set": args.basis_set,
        "solvent": args.solvent,
        "solvation_model": args.solvation_model,
        "solvation_radii": args.solvation_radii,
        "custom_solvation_radii": args.custom_solvation_radii,
        "specifications": args.specifications,
        "freeze": args.freeze,
        "scan": args.scan,
        "sfrom": args.sfrom,
        "sto": args.sto,
        "snsteps": args.snsteps,
        "sstep": args.sstep,
        "density_fitting": args.density_fitting,
        "custom_basis_sets": args.custom_basis_sets,
        "constraints": args.constraints,
        "nproc": args.nproc,
        "mem": args.mem,
        "charge": args.charge,
        "multiplicity": args.mult,
        "parse_name": args.parse_name,
        "trust_me": args.trust_me,
        "d3": args.d3,
        "d3bj": args.d3bj,
        "aux_name": args.aux_name,
        "header": args.header,
        "driver": args.driver,
    }

    if args.preset:
        try:
            preset_params = load_preset(args.preset)
        except InvalidParameter as e:
            warn(str(e))
            exit(0)
        del preset_params["version"]
        for k, v in preset_params.items():
            if k == "specifications":
                params[k] = " ".join([v, params[k]])
            elif params[k] == default_params[k]:
                params[k] = v

    calcs = []
    for ind, (name, xyz, file) in enumerate(zip(names, xyzs, files)):
        if len(outputs) >= ind + 1:
            output = outputs[ind]
        else:
            output = None

        try:
            calc = gen_obj(name=name, xyz=xyz, file=file, output=output, **params)
        except CCInputException as e:
            print(f"!!! {str(e)} !!!")
            exit(0)
        else:
            calcs.append(calc)
    return calcs, outputs



import basis_set_exchange as bse

from ccinput.constants import (
    CalcType,
    ATOMIC_NUMBER,
    LOWERCASE_ATOMIC_SYMBOLS,
    SOFTWARE_BASIS_SETS,
)
from ccinput.utilities import (
    get_method,
    get_basis_set,
    get_solvent,
    get_abs_basis_set,
    clean_xyz,
    warn,
)
from ccinput.exceptions import (
    InvalidParameter,
    UnimplementedError,
    ImpossibleCalculation,
)


class OrcaCalculation:
    calc = None

    has_scan = False
    pal = 0
    blocks = []

    TEMPLATE = """!{}
    *xyz {} {}
    {}*
    {}"""
    # Command Line
    # Charge
    # Multiplicity
    # XYZ structure
    # Options blocks

    command_line = ""
    xyz_structure = ""
    block_lines = ""

    input_file = ""

    CALC_TYPES = [
        CalcType.SP,
        CalcType.OPT,
        CalcType.CONSTR_OPT,
        CalcType.FREQ,
        CalcType.TS,
        CalcType.MEP,
        CalcType.NMR,
        CalcType.UVVIS,
        CalcType.MO,
        CalcType.OPTFREQ,
    ]

    def __init__(self, calc):
        self.calc = calc
        self.has_scan = False
        self.pal = 0
        self.blocks = []
        self.command_line = ""
        self.additional_commands = ""
        self.xyz_structure = ""
        self.block_lines = ""
        self.input_file = ""
        self.specifications = {}
        self.solvation_radii = {}
        self.aux_basis_sets = {}

        if self.calc.type not in self.CALC_TYPES:
            raise ImpossibleCalculation(
                f"ORCA 5 does not support calculations of type {self.calc.type}"
            )

        if self.calc.type == CalcType.UVVIS:
            raise UnimplementedError(
                f"UV-Vis/TD-DFT calculations are currently not interfaced for ORCA 5"
            )
        self.handle_specifications()

        self.handle_command()
        self.handle_custom_basis_sets()
        self.handle_xyz()

        self.handle_pal_mem()
        self.parse_custom_solvation_radii()
        self.handle_solvation()

        self.create_input_file()

    @property
    def confirmed_specifications(self):
        """Returns the effective additional commands; saved in the Parameters object"""
        return self.additional_commands

    def clean(self, s):
        WHITELIST = set(
            "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ/()=-,. "
        )
        return "".join([c for c in s if c in WHITELIST])

    def handle_specifications(self):
        _specifications = (
            self.clean(self.calc.parameters.specifications).lower().strip()
        )

        specifications_list = []
        if _specifications != "":
            sspecs = _specifications.split()
            ind = 0
            while ind < len(sspecs):
                spec = sspecs[ind]
                if spec == "--phirshfeld":
                    HIRSHFELD_BLOCK = """%output
                    Print[ P_Hirshfeld] 1
                    end"""
                    self.blocks.append(HIRSHFELD_BLOCK)
                elif spec == "--nimages":
                    nimages = sspecs[ind + 1]
                    try:
                        nimages = int(nimages)
                    except ValueError:
                        raise InvalidParameter("Invalid specifications")
                    self.specifications["nimages"] = nimages
                    ind += 1
                elif spec[-2:] == "/c":
                    self.aux_basis_sets["C"] = get_basis_set(spec[:-2], "orca")
                elif spec[-3:] == "/jk":
                    self.aux_basis_sets["JK"] = get_basis_set(spec[:-3], "orca")
                elif spec[-2:] == "/j":
                    self.aux_basis_sets["J"] = get_basis_set(spec[:-2], "orca")
                elif spec not in specifications_list:
                    specifications_list.append(spec)

                ind += 1

        if self.calc.parameters.d3:
            specifications_list.append("d3zero")
        elif self.calc.parameters.d3bj:
            specifications_list.append("d3bj")

        if len(specifications_list) > 0:
            self.additional_commands = " ".join(specifications_list)

    def handle_command(self):
        if self.calc.type == CalcType.NMR:
            self.command_line = "NMR "
        elif self.calc.type == CalcType.OPT:
            self.command_line = "OPT "
        elif self.calc.type == CalcType.OPTFREQ:
            self.command_line = "OPT FREQ "
        elif self.calc.type == CalcType.TS:
            self.command_line = "OPTTS "
            if self.calc.parameters.theory_level != "xtb":
                self.blocks.append("%geom\nCalc_Hess true\nend")
        elif self.calc.type == CalcType.MO:
            self.command_line = "SP "
            struct = clean_xyz(self.calc.xyz)

            electrons = 0
            for line in struct.split("\n"):
                if line.strip() == "":
                    continue
                el = line.split()[0]
                electrons += ATOMIC_NUMBER[el]

            electrons -= self.calc.charge

            if self.calc.multiplicity != 1:
                raise InvalidParameter("Unimplemented multiplicity")

            n_HOMO = int(electrons / 2) - 1
            n_LUMO = int(electrons / 2)
            n_LUMO1 = int(electrons / 2) + 1
            n_LUMO2 = int(electrons / 2) + 2

            mo_block = f"""%plots
                        dim1 45
                        dim2 45
                        dim3 45
                        min1 0
                        max1 0
                        min2 0
                        max2 0
                        min3 0
                        max3 0
                        Format Gaussian_Cube
                        MO("in-HOMO.cube",{n_HOMO},0);
                        MO("in-LUMO.cube",{n_LUMO},0);
                        MO("in-LUMOA.cube",{n_LUMO1},0);
                        MO("in-LUMOB.cube",{n_LUMO2},0);
                        end
                        """
            self.blocks.append(mo_block)
        elif self.calc.type == CalcType.FREQ:
            self.command_line = "FREQ "
        elif self.calc.type == CalcType.CONSTR_OPT:
            self.command_line = "OPT "

            if len(self.calc.constraints) == 0:
                raise InvalidParameter("No constraints for constrained optimisation")

            scans = []
            freeze = []

            for constr in self.calc.constraints:
                if constr.scan:
                    scans.append(constr.to_orca())
                else:
                    freeze.append(constr.to_orca())

            if len(scans) > 0:
                scan_block = """%geom Scan
                {}
                end
                end"""
                self.blocks.append(scan_block.format("".join(scans).strip()))

            if len(freeze) > 0:
                freeze_block = """%geom Constraints
                {}
                end
                end"""
                self.blocks.append(freeze_block.format("".join(freeze).strip()))
        elif self.calc.type == CalcType.SP:
            self.command_line = "SP "
        elif self.calc.type == CalcType.MEP:  #### Second structure to handle
            self.command_line = "NEB "
            neb_block = """%neb
                        product "{}.xyz"
                        nimages {}
                        end"""
            if "nimages" in self.specifications:
                nimages = self.specifications["nimages"]
            else:
                nimages = 8
            self.blocks.append(neb_block.format(self.calc.aux_name, nimages))

        method = get_method(self.calc.parameters.method, "orca")
        if self.calc.parameters.theory_level not in [
            "xtb",
            "semiempirical",
            "special",
        ]:
            basis_set = get_basis_set(self.calc.parameters.basis_set, "orca")
            self.command_line += f"{method} {basis_set} "
            if self.calc.parameters.theory_level in ["mp2", "cc"] and (
                method.find("RI") != -1 or method.find("LPNO") != -1
            ):
                # If we need an auxiliary basis set
                if "C" not in self.aux_basis_sets:
                    warn(f"No C auxiliary basis set specified, using {basis_set}")
                    self.aux_basis_sets["C"] = basis_set
        else:
            self.command_line += f"{method} "

        for aux_t, aux_bs in self.aux_basis_sets.items():
            self.command_line += f"{aux_bs}/{aux_t} "

    def handle_custom_basis_sets(self):
        if len(self.calc.parameters.custom_basis_sets) == 0:
            return

        unique_atoms = []
        for line in self.calc.xyz.split("\n"):
            if line.strip() == "":
                continue
            a, *_ = line.split()
            if a not in unique_atoms:
                unique_atoms.append(a)

        BS_TEMPLATE = """%basis
        {}
        end"""

        custom_bs = ""
        for el, bs_keyword in self.calc.parameters.custom_basis_sets.items():
            if el not in unique_atoms:
                continue

            try:
                el_num = ATOMIC_NUMBER[el]
            except KeyError:
                raise InvalidParameter("Invalid atom in custom basis set string")

            abs_keyword = get_abs_basis_set(bs_keyword)
            success = False
            if abs_keyword in SOFTWARE_BASIS_SETS["orca"]:
                _custom_bs = (
                    f'NewGTO {el} "{SOFTWARE_BASIS_SETS["orca"][abs_keyword]}" end\n'
                )
                gbs = bse.get_basis(bs_keyword, elements=[el_num])
                if "ecp_potentials" in gbs["elements"][str(el_num)]:
                    # Search for the right ECP
                    hits = bse.filter_basis_sets(family=gbs["family"], role="orbital")
                    ecp_keyword = ""
                    for name, hit in hits.items():
                        if hit["function_types"] == ["scalar_ecp"]:
                            ecp_keyword = get_basis_set(name, "orca")
                            break
                    else:
                        warn(
                            "Could not find the name of the ECP linked to {SOFTWARE_BASIS_SETS['orca'][abs_keyword]}, adding manually..."
                        )

                    if ecp_keyword:
                        _custom_bs += f'NewECP {el} "{ecp_keyword}" end\n'
                        success = True
                else:
                    success = True

                # TODO: handle auxiliary basis sets

            if success:
                custom_bs += _custom_bs.strip()
            else:
                bs = bse.get_basis(
                    bs_keyword, fmt="ORCA", elements=[el_num], header=False
                ).strip()
                sbs = bs.split("\n")
                if bs.find("ECP") != -1:
                    clean_bs = "\n".join(sbs[3:]).strip() + "\n"
                    clean_bs = clean_bs.replace("\n$END", "$END").replace("$END", "end")
                    custom_bs += f"newgto {el}\n"
                    custom_bs += clean_bs.strip()
                else:
                    clean_bs = "\n".join(sbs[3:-1]).strip() + "\n"
                    custom_bs += f"newgto {el}\n"
                    custom_bs += clean_bs.strip()
                    custom_bs += "end"

        if custom_bs != "":
            self.blocks.append(BS_TEMPLATE.format(custom_bs))

    def handle_xyz(self):
        lines = [i + "\n" for i in clean_xyz(self.calc.xyz).split("\n") if i != ""]
        self.xyz_structure = "".join(lines)

    def handle_pal_mem(self):
        if self.calc.parameters.theory_level == "semiempirical":
            self.pal = 1
            self.mem_per_core = self.calc.mem
        else:
            self.pal = self.calc.nproc
            self.mem_per_core = int(self.calc.mem / self.calc.nproc)

        pal_mem_block = f"""%MaxCore {self.mem_per_core}
        %pal
        nprocs {self.pal}
        end"""

        self.blocks.append(pal_mem_block)

    def parse_custom_solvation_radii(self):
        for radius in self.calc.parameters.custom_solvation_radii.split(";"):
            if radius.strip() == "":
                continue
            sradius = radius.split("=")
            if len(sradius) != 2:
                raise InvalidParameter(
                    f"Invalid custom solvation radius specification: '{radius}': must follow the pattern '<atom1>=<radius1>;...'"
                )

            element, rad = sradius
            if element not in LOWERCASE_ATOMIC_SYMBOLS:
                raise InvalidParameter(
                    f"Invalid element in custom solvation radius specifications: '{element}'"
                )

            _element = ATOMIC_NUMBER[LOWERCASE_ATOMIC_SYMBOLS[element]]

            try:
                _rad = float(rad)
            except ValueError:
                raise InvalidParameter(
                    f"Invalid custom solvation radius for element {element}: '{rad}'"
                )
            self.solvation_radii[_element] = _rad

    def get_radii_specs(self):
        radii_specs = ""
        for el, rad in self.solvation_radii.items():
            radii_specs += f"radius[{el}] {rad:.2f}\n"
        return radii_specs

    def handle_solvation(self):
        if self.calc.parameters.solvent.lower() not in ["vacuum", ""]:
            solvent_keyword = get_solvent(
                self.calc.parameters.solvent,
                self.calc.parameters.software,
                solvation_model=self.calc.parameters.solvation_model,
            )
            model = self.calc.parameters.solvation_model
            radii_set = self.calc.parameters.solvation_radii
            custom_radii = self.calc.parameters.custom_solvation_radii

            solv_block = ""

            if self.calc.parameters.method[:3].lower() == "xtb":
                self.command_line += f" ALPB({solvent_keyword})"
            elif model == "smd":
                # Refined solvation radii
                # E. Engelage, N. Schulz, F. Heinen, S. M. Huber, D. G. Truhlar,
                # C. J. Cramer, Chem. Eur. J. 2018, 24, 15983–15987.
                if radii_set == "smd18":
                    if 53 not in self.solvation_radii:
                        self.solvation_radii[53] = 2.74
                    if 35 not in self.solvation_radii:
                        self.solvation_radii[35] = 2.60

                radii_specs = self.get_radii_specs()
                solv_block = f"""%cpcm
                smd true
                SMDsolvent "{solvent_keyword}"
                {radii_specs}end"""
                self.blocks.append(solv_block)
            elif model == "cpcm":
                self.command_line += f"CPCM({solvent_keyword}) "
                radii_specs = self.get_radii_specs()
                if radii_specs != "":
                    solv_block = f"""%cpcm
                    {radii_specs}end"""
                    self.blocks.append(solv_block)
                if radii_set not in ["", "default", "bondi"]:
                    raise UnimplementedError(
                        "As of now, ccinput does not know how to request "
                        "other solvation radii than the default ones in ORCA with CPCM"
                    )
            else:
                raise InvalidParameter(
                    f"Invalid solvation model for ORCA: "
                    f"'{self.calc.parameters.solvation_model}'"
                )

    def create_input_file(self):
        self.block_lines = "\n".join(self.blocks)
        cmd = f"{self.command_line} {self.additional_commands}".replace("  ", " ")
        raw = self.TEMPLATE.format(
            cmd,
            self.calc.charge,
            self.calc.multiplicity,
            self.xyz_structure,
            self.block_lines,
        )
        self.input_file = "\n".join([i.strip() for i in raw.split("\n")])

    @property
    def output(self):
        return self.input_file



from unittest import TestCase

from ccinput.utilities import standardize_xyz
from ccinput.exceptions import InvalidXYZ


class XyzTests(TestCase):
    def test_standard_xyz(self):
        xyz = "1\n\nCl 0.0 0.0 0.0\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_headerless_xyz(self):
        xyz = "Cl 0.0 0.0 0.0\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_extra_line(self):
        xyz = "1\n\nCl 0.0 0.0 0.0\n\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_extra_lines(self):
        xyz = "1\n\nCl 0.0 0.0 0.0\n\n\n\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_invalid_header(self):
        xyz = "2\n\nCl 0.0 0.0 0.0\n"
        with self.assertRaises(InvalidXYZ):
            standardize_xyz(xyz)

    def test_invalid_header_space(self):
        xyz = "2\n\nCl 0.0 0.0 0.0\n\n"
        with self.assertRaises(InvalidXYZ):
            standardize_xyz(xyz)

    def test_extra_spacing(self):
        xyz = """1
        header
        Cl 0.0 0.0 0.0


        """
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_multiple_atoms(self):
        xyz = """3
        header
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        """
        self.assertEqual(
            standardize_xyz(xyz),
            "Cl   0.00000000   0.00000000   0.00000000\nCl   0.00000000   0.00000000   0.00000000\nCl   0.00000000   0.00000000   0.00000000\n",
        )

    def test_multiple_atoms_invalid_header(self):
        xyz = """2
        header
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        """
        with self.assertRaises(InvalidXYZ):
            standardize_xyz(xyz)

    def test_multiple_atoms_invalid_header2(self):
        xyz = """4
        header
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        Cl 0.0 0.0 0.0
        """
        with self.assertRaises(InvalidXYZ):
            standardize_xyz(xyz)

    def test_invalid_element(self):
        xyz = "1\n\nBl 0.0 0.0 0.0\n"
        with self.assertRaises(InvalidXYZ):
            standardize_xyz(xyz)

    def test_lowercase(self):
        xyz = "1\n\ncl 0.0 0.0 0.0\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )

    def test_uppercase(self):
        xyz = "1\n\nCL 0.0 0.0 0.0\n"
        self.assertEqual(
            standardize_xyz(xyz), "Cl   0.00000000   0.00000000   0.00000000\n"
        )



import hashlib
import os
from itertools import zip_longest

from ccinput.exceptions import (
    InvalidParameter,
    InternalError,
    ImpossibleCalculation,
    MissingParameter,
)
from ccinput.utilities import (
    get_abs_software,
    get_method,
    get_abs_basis_set,
    get_abs_solvent,
    get_theory_level,
    standardize_memory,
    get_npxyz,
    get_coord,
    has_dispersion_parameters,
    warn,
    indexify,
    get_charge_mult_from_name,
)
from ccinput.constants import ATOMIC_NUMBER, SYN_SOFTWARE, BASIS_SET_EXCHANGE_KEY


class Calculation:
    """
    Holds all the data required to generate an input file.
    Its fields are the parameters likely to change (charge, multiplicity,
    xyz, calculation type...). The other parameters are contained in the
    Parameters class (accessed through self.parameters).
    """

    def __init__(
        self,
        xyz,
        parameters,
        type,
        constraints=[],
        nproc=1,
        mem=1000,
        charge=0,
        multiplicity=1,
        parse_name=False,
        aux_name=None,
        name=None,
        header="File created by ccinput",
        software="",
        file=None,
        driver="none",
        **kwargs,
    ):
        self.xyz = xyz
        self.parameters = parameters
        self.type = type
        self.file = file

        if software not in SYN_SOFTWARE:
            raise InvalidParameter(f"Invalid software: '{software}'")

        if nproc == 0:
            raise MissingParameter("Number of cores unspecified")

        try:
            self.nproc = int(nproc)
        except ValueError:
            raise InvalidParameter(f"Invalid number of cores: '{nproc}'")

        if abs(self.nproc - float(nproc)) > 1e-4:
            raise InvalidParameter(
                f"The number of cores must be an integer (received '{nproc}')"
            )

        if self.nproc < 1:
            raise InvalidParameter(
                f"The number of cores must at least 1 (received '{nproc}')"
            )

        try:
            self.mem = standardize_memory(mem)
        except InvalidParameter:
            if software in ["gaussian"]:
                raise

        try:
            self.charge = int(charge)
        except ValueError:
            raise InvalidParameter(f"Invalid charge: '{charge}'")

        if abs(self.charge - float(charge)) > 1e-4:
            raise InvalidParameter(f"Charge must be an integer (received '{charge}')")

        try:
            self.multiplicity = int(multiplicity)
        except ValueError:
            raise InvalidParameter(f"Invalid multiplicity: '{multiplicity}'")

        if abs(self.multiplicity - float(multiplicity)) > 1e-4:
            raise InvalidParameter(
                f"Multiplicity must be an integer (received '{multiplicity}')"
            )
        if self.multiplicity < 1:
            raise InvalidParameter(
                f"Multiplicity must at least 1 (received '{multiplicity}')"
            )

        if parse_name:
            if not file:
                raise InvalidParameter(
                    f"Cannot parse the charge and multiplicity from the file name: no file given"
                )

            self.charge, self.multiplicity = get_charge_mult_from_name(file)

        self.verify_charge_mult()
        self.constraints = constraints

        if name is None:
            if file is None:
                self.name = "calc"
            else:
                self.name = os.path.splitext(os.path.basename(file))[0]
        else:
            self.name = name

        if aux_name is None:
            self.aux_name = "calc2"
        else:
            self.aux_name = aux_name
        self.header = header
        self.driver = driver
        self.kwargs = kwargs

    def verify_charge_mult(self):
        """Verifies that the requested charge and multiplicity are possible for the structure"""
        electrons = 0
        for line in self.xyz.split("\n"):
            if line.strip() == "":
                continue
            el = line.split()[0]
            electrons += ATOMIC_NUMBER[el]

        electrons -= self.charge
        odd_e = electrons % 2
        odd_m = self.multiplicity % 2

        if odd_e == odd_m:
            raise ImpossibleCalculation(
                f"This combination of charge ({self.charge}) "
                + f"and multiplicity ({self.multiplicity}) is impossible"
            )


class Parameters:
    """
    Holds all the parameters about the computational method.
    These parameters do not depend on the particular system
    (e.g. regarding the charge and multiplicity) and can be reused.
    """

    def __init__(
        self,
        software,
        solvent="",
        solvation_model="",
        solvation_radii="",
        custom_solvation_radii="",
        basis_set="",
        method="",
        specifications="",
        density_fitting="",
        custom_basis_sets="",
        d3=False,
        d3bj=False,
        trust_me=False,
        **kwargs,
    ):
        self.solvent = get_abs_solvent(solvent, trust_me=trust_me)
        if self.solvent != "":
            self.solvation_model = solvation_model.lower()
            self.solvation_radii = solvation_radii.lower()
            self.custom_solvation_radii = custom_solvation_radii.lower()

            if self.solvation_model.strip() == "":
                raise InvalidParameter(
                    "No solvation model specified, " + "although solvation is requested"
                )
            if self.solvation_radii.strip() == "":
                warn("No solvation radii specified; using default radii")
        else:
            self.solvation_model = ""
            self.solvation_radii = ""
            self.custom_solvation_radii = ""

        self.software = get_abs_software(software)

        if method == "":
            if "functional" in kwargs:
                self.method = get_method(kwargs["functional"], self.software)
            elif self.software == "xtb":
                self.method = "gfn2-xtb"
            else:
                raise InvalidParameter("No calculation method specified (method='...')")

        else:
            self.method = get_method(method, self.software)

        self.theory_level = get_theory_level(self.method)

        if self.theory_level not in ["semiempirical", "xtb", "special"]:
            self.basis_set = get_abs_basis_set(basis_set, trust_me=trust_me)
        else:
            self.basis_set = ""

        if d3 and d3bj:
            raise InvalidParameter(
                "Cannot use both D3(0) and D3BJ dispersion corrections"
            )

        self.d3 = d3
        self.d3bj = d3bj

        if d3 and not has_dispersion_parameters(self.method, version="d3"):
            warn(
                f"Your calculation requests a method ({self.method}) that may not have D3 "
                "parameters. Be aware that the calculation might result in an error"
            )

        if d3bj and not has_dispersion_parameters(self.method, version="d3bj"):
            warn(
                f"Your calculation requests a method ({self.method}) that may not have D3BJ "
                "parameters. Be aware that the calculation might result in an error"
            )

        self.specifications = specifications
        self.density_fitting = density_fitting

        dict_cbs = {}

        for _entry in custom_basis_sets.split(";"):
            entry = _entry.replace(" ", "")
            if entry == "":
                continue
            sentry = entry.split("=")

            if len(sentry) != 2:
                raise InvalidParameter(f"Invalid custom basis set string: '{entry}'")

            el, bs_str = sentry

            if el in dict_cbs:
                raise InvalidParameter(
                    f"The element {el} was specified more than once in the custom basis set string"
                )

            bs_index = indexify(bs_str)

            if bs_index in BASIS_SET_EXCHANGE_KEY:
                bs_keyword = BASIS_SET_EXCHANGE_KEY[bs_index]
            else:
                bs_keyword = get_abs_basis_set(bs_str, trust_me=trust_me)

            dict_cbs[el] = bs_keyword

        self.custom_basis_sets = dict_cbs
        self.kwargs = kwargs

    def __eq__(self, other):
        values = [(k, v) for k, v in self.__dict__.items()]
        other_values = [(k, v) for k, v in other.__dict__.items()]

        return values == other_values

    @property
    def md5(self):
        """
        Returns a hash digest of the parameters to easily compare
        parameters without using objects.
        """
        values = [(k, v) for k, v in self.__dict__.items()]
        params_str = ""
        for k, v in values:
            if isinstance(v, int):
                params_str += f"{k}={v};"
            elif isinstance(v, str):
                params_str += f"{k}={v.lower()};"
            else:
                raise InternalError("Unknown value type")
        hash = hashlib.md5(bytes(params_str, "UTF-8"))
        return hash.digest()


class Constraint:
    """
    Class to contain a single constraint (freeze, scan and the like) and provide useful functions for it.
    """

    def __init__(
        self,
        scan=False,
        start_d=None,
        end_d=None,
        step_size=None,
        num_steps=None,
        ids=[],
        xyz=None,
        software="",
    ):
        self.scan = scan
        self.ids = ids  # One-indexed (Gaussian like, not like ORCA)
        self.software = software

        if len(ids) < 2 or len(ids) > 4:
            raise InvalidParameter(
                f"Invalid number of atoms: {len(ids)}, needs to be between 2 and 4"
            )

        _xyz = get_npxyz(xyz)

        if max(ids) > len(_xyz):
            raise InvalidParameter(
                f"Invalid atom index {max(ids)}: larger than the number of atoms"
            )
        if len(ids) != len(set(ids)):
            raise InvalidParameter(
                f"The provided atom indices are not all unique: {ids}"
            )

        if self.scan:
            num_params = 0
            for p in [end_d, step_size, num_steps]:
                if p is not None:
                    num_params += 1

            if num_params < 2:
                raise InvalidParameter("Not enough constraint parameters given")
            elif num_params == 3:
                raise InvalidParameter(
                    "Too many constraint parameters given, more than one possibility of unique parameters"
                )

            if start_d is not None and end_d is not None and xyz is None:
                raise InvalidParameter(
                    "The XYZ structure needs to be specified in order to calculate the initial coordinate value"
                )

        if software != "gaussian":
            self.start_d = start_d
        else:
            if start_d is not None:
                warn(
                    "Gaussian only allows scans that start from the current structure; "
                    + "overriding the specified initial value"
                )
            self.start_d = None

        self.end_d = end_d
        self.step_size = step_size
        self.num_steps = num_steps

        if self.start_d is None:
            self.start_d = get_coord(_xyz, self.ids)

        if self.scan:
            self.complete_parameters()

    def complete_parameters(self):
        if self.end_d is None:
            assert self.step_size is not None
            assert self.num_steps is not None

            self.end_d = self.start_d + self.num_steps * self.step_size

        if self.step_size is None:
            assert self.end_d is not None
            assert self.num_steps is not None

            self.step_size = round((self.end_d - self.start_d) / self.num_steps, 2)

        if self.num_steps is None:
            assert self.end_d is not None
            assert self.step_size is not None

            # Make sure the number of steps is positive overall
            # Also use the absolute step size here, as the start and end are defined
            # An mathematically incorrect sign won't matter, as the step size is recalculated
            self.num_steps = abs(int((self.end_d - self.start_d) / abs(self.step_size)))

            # Adjust the step size to end up exactly at the end point
            self.step_size = round((self.end_d - self.start_d) / self.num_steps, 2)

    def to_orca(self):
        ids_str = " ".join([str(i - 1) for i in self.ids])
        type = len(self.ids)

        if type == 2:
            t = "B"
        elif type == 3:
            t = "A"
        elif type == 4:
            t = "D"

        if self.scan:
            return f"{t} {ids_str} = {self.start_d:.2f}, {self.end_d:.2f}, {self.num_steps}\n"
        else:
            return f"{{ {t} {ids_str} C }}\n"

    def to_gaussian(self):
        ids_str = " ".join([str(i) for i in self.ids])
        type = len(self.ids)

        if type == 2:
            t = "B"
        elif type == 3:
            t = "A"
        elif type == 4:
            t = "D"
            if self.scan:
                # Different sign convention used by Gaussian for dihedral angles (?)
                self.step_size *= -1

        if self.scan:
            return f"{t} {ids_str} S {self.num_steps} {self.step_size}\n"
        else:
            return f"{t} {ids_str} F\n"

    def to_xtb(self):
        ids_str = ", ".join(
            [str(i) for i in self.ids] + ["auto"]
        )  # Maybe add support for other starting points in the future
        type = len(self.ids)

        if type == 2:
            t = "distance"
        elif type == 3:
            t = "angle"
        elif type == 4:
            t = "dihedral"

        return f"{t}: {ids_str}\n"


def parse_freeze_constraints(arr, xyz_str, software=""):
    if len(arr) == 0:
        return []
    constr = ""
    for c in arr:
        constr += f'Freeze/{"_".join([str(i) for i in c])};'

    return parse_str_constraints(constr, xyz_str, software=software)


def parse_scan_constraints(arr, sfrom, sto, snsteps, sstep, xyz_str, software=""):
    if len(arr) == 0:
        return []

    scans = []
    for ids, fro, to, nsteps, step in zip_longest(arr, sfrom, sto, snsteps, sstep):
        if ids is None:
            raise InvalidParameter(
                "Not enough sets of atom indices specified for the number of other parameters"
            )
        _ids = [int(i) for i in ids]
        scans.append(
            gen_constraint(
                _ids,
                xyz_str,
                "scan",
                start_str=fro,
                end_str=to,
                nsteps_str=nsteps,
                step_str=step,
                software=software,
            )
        )

    return scans


def parse_str_constraints(s, xyz_str, software=""):
    if s.strip() == "":
        return []

    if software == "":
        warn(
            "No software specified for the constraints; the behaviour might be incorrect"
        )

    constraints = []
    cs = s.split(";")
    for c in cs:
        if c.strip() == "":
            continue

        if c.count("/") != 1:
            raise InvalidParameter(f"Invalid constraint string: '{c}'")

        specs_str, ids_str = c.lower().split("/")

        try:
            ids = [int(i) for i in ids_str.split("_")]
        except ValueError:
            raise InvalidParameter(
                f"Could not parse the atom numbers from the string '{ids_str}'"
            )

        constraints.append(
            gen_constraint(ids, xyz_str, *specs_str.split("_"), software=software)
        )

    return constraints


def gen_constraint(
    ids,
    xyz_str,
    option,
    start_str=None,
    end_str=None,
    nsteps_str=None,
    step_str=None,
    software="",
):
    """
    Generate a constraint object from arrays of parameters.
    In case of multiple values in the arrays, each parameter with the same index
    is assumed to be of the same command constraint.

    The atom indices must already be integers in ids as array of arrays of ints.

    """
    t = len(ids)

    if option not in ["scan", "freeze"]:
        raise InvalidParameter(f"Invalid type of scan: '{specs[0]}'")

    if option == "scan":
        if start_str is not None:
            if isinstance(start_str, str) and start_str.lower() == "auto":
                start_d = None
            else:
                try:
                    start_d = float(start_str)
                except ValueError:
                    raise InvalidParameter(f"Invalid initial value: '{start_str}'")

                if t == 2 and start_d < 0.01:
                    raise InvalidParameter(f"Invalid initial distance: '{start_str}'")
        else:
            start_d = None

        if end_str is not None:
            try:
                end_d = float(end_str)
            except ValueError:
                raise InvalidParameter(f"Invalid final value: '{end_str}'")

            if t == 2 and end_d < 0.01:
                raise InvalidParameter(f"Invalid final distance: '{end_str}'")
        else:
            end_d = None

        if nsteps_str is not None:
            try:
                num_steps = int(nsteps_str)
            except ValueError:
                raise InvalidParameter(f"Invalid number of steps: '{nsteps_str}'")

            if num_steps < 1:
                raise InvalidParameter(f"Invalid number of steps: '{nsteps_str}'")
        else:
            num_steps = None

        if step_str is not None:
            try:
                step_size = float(step_str)
            except ValueError:
                raise InvalidParameter(f"Invalid final value: '{step_str}'")
        else:
            step_size = None

    else:
        start_d = None
        end_d = None
        num_steps = None
        step_size = None

    return Constraint(
        scan=option == "scan",
        start_d=start_d,
        end_d=end_d,
        num_steps=num_steps,
        step_size=step_size,
        ids=ids,
        xyz=xyz_str,
        software=software.lower(),
    )



from pathlib import Path
from setuptools import setup, find_packages
from versioneer import get_version, get_cmdclass

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="ccinput",
    version=get_version(),
    cmdclass=get_cmdclass(),
    description="Computational Chemistry Input Generator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/cyllab/ccinput",
    author="Raphaël Robidas",
    author_email="Raphael.Robidas@USherbrooke.ca",
    license="BSD 3-Clause",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "ccinput = ccinput.wrapper:cmd",
        ],
    },
    install_requires=[
        "periodictable",
        "basis_set_exchange",
        "numpy",
        "versioneer",
        "appdirs",
    ],
    test_suite="nose.collector",
    tests_require=[
        "nose",
        "mock",
    ],
    python_requires=">=3.6",
    zip_safe=False,
)



import shlex
import os
from mock import patch
import tempfile
import json

from ccinput.calculation import Calculation, Parameters
from ccinput.wrapper import gen_input, get_input_from_args, get_parser, cmd
from ccinput.tests.testing_utilities import InputTests
from ccinput import presets

from contextlib import contextmanager, redirect_stdout
from os import devnull

PRESET_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "presets")


@contextmanager
def hide_cmd_output():
    with open(devnull, "w") as gobble:
        with redirect_stdout(gobble) as out:
            yield out


class CliEquivalenceTests(InputTests):
    def test_basic(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_no_nproc(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' --mem 1G -c -1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_basic_file(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "name": "ethanol",
        }
        line = f"gaussian sp HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_freeze_distance(self):
        args = {
            "software": "gaussian",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Freeze/1_2;",
            "name": "ethanol",
        }
        line = f"gaussian constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --freeze 1 2"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_freeze_angle(self):
        args = {
            "software": "gaussian",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Freeze/3_2_4;",
            "name": "ethanol",
        }
        line = f"gaussian constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --freeze 3 2 4"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_freeze_dihedral(self):
        args = {
            "software": "gaussian",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Freeze/6_3_1_2;",
            "name": "ethanol",
        }
        line = f"gaussian constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --freeze 6 3 1 2"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_distance(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 --from 2.0 --to 1.0 --nsteps 10"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_distance_multiple(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2;Scan_2.0_1.0_10/3_4;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 --from 2.0 --to 1.0 --nsteps 10 --scan 3 4 --from 2.0 --to 1.0 --nsteps 10"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_distance_multiple_no_from(self):
        args = {
            "software": "gaussian",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2;Scan_2.0_1.0_10/3_4;",
            "name": "ethanol",
        }
        line = f"gaussian constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 --to 1.0 --nsteps 10 --scan 3 4 --to 1.0 --nsteps 10"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_distance_multiple_step(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2;Scan_2.0_1.0_10/3_4;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 --from 2.0 --to 1.0 --step -0.1 --scan 3 4 --from 2.0 --to 1.0 --step -0.1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_distance_step_wrong_sign(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2;Scan_2.0_1.0_10/3_4;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 --from 2.0 --to 1.0 --step 0.1 --scan 3 4 --from 2.0 --to 1.0 --step 0.1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_angle(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2_3;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 3 --from 2.0 --to 1.0 --nsteps 10"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_scan_dihedral(self):
        args = {
            "software": "orca",
            "type": "constr_opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "constraints": "Scan_2.0_1.0_10/1_2_3_4;",
            "name": "ethanol",
        }
        line = f"orca constr_opt HF -bs Def2SVP -f {self.struct('ethanol')} -n 1 --mem 1G --scan 1 2 3 4 --from 2.0 --to 1.0 --nsteps 10"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_name(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "name": "Chloride in vacuum",
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --name 'Chloride in vacuum'"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_custom_basis_set(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "custom_basis_sets": "Cl=Def2-SVPD;",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 -cbs 'Cl=Def2-SVPD;'"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_solvation(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "solvation_model": "SMD",
            "solvent": "Chloroform",
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --solvent chloroform --solvation_model smd"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_solvation_smd18(self):
        args = {
            "software": "gaussian",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "solvation_model": "SMD",
            "solvent": "Chloroform",
            "solvation_radii": "SMD18",
        }
        line = "gaussian sp HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --solvent chloroform --solvation_model smd --solvation_radii SMD18"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_opt_freq(self):
        args = {
            "software": "gaussian",
            "type": "opt+freq",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }
        line = "gaussian opt+freq HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_opt_freq2(self):
        args = {
            "software": "gaussian",
            "type": "opt+freq",
            "method": "HF",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }
        line = "gaussian opt-freq HF -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_d3_gaussian(self):
        args = {
            "software": "gaussian",
            "type": "opt",
            "method": "M06",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "d3": True,
        }
        line = "gaussian opt M06 -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --d3"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_d3bj_gaussian(self):
        args = {
            "software": "gaussian",
            "type": "opt",
            "method": "PBE0",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "d3bj": True,
        }
        line = (
            "gaussian opt PBE0 -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --d3bj"
        )
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_d3_orca(self):
        args = {
            "software": "ORCA",
            "type": "opt",
            "method": "M06",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "d3": True,
        }
        line = "orca opt M06 -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --d3"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_d3bj_orca(self):
        args = {
            "software": "ORCA",
            "type": "opt",
            "method": "PBE0",
            "basis_set": "Def2SVP",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "d3bj": True,
        }
        line = "orca opt PBE0 -bs Def2SVP --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 --d3bj"
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_winchars(self):
        presets.data_dir = PRESET_DIR
        args = {
            "software": "gaussian",
            "type": "opt",
            "method": "HF",
            "basis_set": "def2tzvp",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }

        line = '--preset winchars --xyz "Cl 0 0 0" -c -1'
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_whitespace(self):
        presets.data_dir = PRESET_DIR
        args = {
            "software": "gaussian",
            "type": "opt",
            "method": "HF",
            "basis_set": "def2tzvp",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
        }

        line = '--preset whitespace --xyz "Cl 0 0 0" -c -1'
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_preset_additive_specifications(self):
        presets.data_dir = PRESET_DIR
        args = {
            "software": "gaussian",
            "type": "opt",
            "method": "HF",
            "basis_set": "def2tzvp",
            "xyz": "Cl 0 0 0\n",
            "nproc": 1,
            "mem": "1G",
            "charge": -1,
            "specifications": "nosymm 5d",
        }

        line = '--preset specification --xyz "Cl 0 0 0" -c -1 --specification "5d"'
        self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_sp_xtb(self):
        args = {
            "software": "xtb",
            "type": "sp",
            "file": self.struct("ethanol"),
        }

        inp = gen_input(**args)

        line = "xtb ethanol.xyz"
        self.assertEqual(inp, line)

    def test_opt_xtb(self):
        args = {
            "software": "xtb",
            "type": "opt",
            "file": self.struct("ethanol"),
        }

        inp = gen_input(**args)

        line = "xtb ethanol.xyz --opt tight"
        self.assertEqual(inp, line)


class ManualCliTests(InputTests):
    def setUp(self):
        self.warnings = []

    def get_warn(self, msg):
        self.warnings.append(msg)

    def test_multiple_files_no_output(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 0)

        args1 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
        }
        args2 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("CH4"),
            "nproc": 1,
            "mem": "1G",
        }

        inp1 = gen_input(**args1)
        inp2 = gen_input(**args2)

        self.assertTrue(self.is_equivalent(inp1, objs[0].input_file))
        self.assertTrue(self.is_equivalent(inp2, objs[1].input_file))

    def test_multiple_files_output(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -o test.inp -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)

        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 2)

        self.assertEqual(os.path.basename(outputs[0]), "test_ethanol.inp")
        self.assertEqual(os.path.basename(outputs[1]), "test_CH4.inp")

        args1 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
        }
        args2 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("CH4"),
            "nproc": 1,
            "mem": "1G",
        }

        inp1 = gen_input(**args1)
        inp2 = gen_input(**args2)

        self.assertTrue(self.is_equivalent(inp1, objs[0].input_file))
        self.assertTrue(self.is_equivalent(inp2, objs[1].input_file))

    def test_multiple_files_output_no_name(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -o .inp -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 2)

        self.assertEqual(outputs[0], "ethanol.inp")
        self.assertEqual(outputs[1], "CH4.inp")

    def test_multiple_files_output_name_no_override(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -o .inp -n 1 --mem 1G --name test"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 2)

        self.assertEqual(outputs[0], "ethanol.inp")
        self.assertEqual(outputs[1], "CH4.inp")

    def test_multiple_files_output_directory(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -o calc_dir/.inp -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 2)

        self.assertEqual(outputs[0], "calc_dir/ethanol.inp")
        self.assertEqual(outputs[1], "calc_dir/CH4.inp")

    def test_multiple_files_specifications_orca(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -n 1 --mem 1G --specifications 'tightscf'"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 0)

        args1 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "specifications": "tightscf",
        }
        args2 = {
            "software": "ORCA",
            "type": "sp",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("CH4"),
            "nproc": 1,
            "mem": "1G",
            "specifications": "tightscf",
        }

        inp1 = gen_input(**args1)
        inp2 = gen_input(**args2)

        self.assertTrue(self.is_equivalent(inp1, objs[0].input_file))
        self.assertTrue(self.is_equivalent(inp2, objs[1].input_file))

    def test_multiple_files_specifications_gaussian(self):
        cmd_line = f"g16 opt HF -bs Def2SVP -f {self.struct('ethanol')} {self.struct('CH4')} -n 1 --mem 1G --specifications 'opt(maxstep=5)'"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 2)
        self.assertEqual(len(outputs), 0)

        args1 = {
            "software": "g16",
            "type": "opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("ethanol"),
            "nproc": 1,
            "mem": "1G",
            "name": "ethanol",
            "specifications": "opt(maxstep=5)",
        }
        args2 = {
            "software": "g16",
            "type": "opt",
            "method": "HF",
            "basis_set": "Def2SVP",
            "file": self.struct("CH4"),
            "nproc": 1,
            "mem": "1G",
            "name": "CH4",
            "specifications": "opt(maxstep=5)",
        }

        inp1 = gen_input(**args1)
        inp2 = gen_input(**args2)

        self.assertTrue(self.is_equivalent(inp1, objs[0].input_file))
        self.assertTrue(self.is_equivalent(inp2, objs[1].input_file))

    def test_single_file_output(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} -o calc_dir/ethanol.inp -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 1)
        self.assertEqual(len(outputs), 1)

        self.assertEqual(outputs[0], "calc_dir/ethanol.inp")

    def test_single_file_output_change_filename(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} -o calc_dir/sp_HF.inp -n 1 --mem 1G"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 1)
        self.assertEqual(len(outputs), 1)

        self.assertEqual(outputs[0], "calc_dir/sp_HF.inp")

    def test_single_file_output_change_name(self):
        cmd_line = f"orca sp HF -bs Def2SVP -f {self.struct('ethanol')} -o calc_dir/sp_HF.inp -n 1 --mem 1G --name my_sp"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 1)
        self.assertEqual(len(outputs), 1)

        self.assertEqual(outputs[0], "calc_dir/sp_HF.inp")

    def test_xtb_commandline(self):
        cmd_line = f"xtb sp gfn2-xtb -f {self.struct('ethanol')}"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 1)
        self.assertEqual(len(outputs), 0)

        self.assertEqual(objs[0].input_file, "")
        self.assertEqual(objs[0].command, "xtb ethanol.xyz")

    def test_crest_commandline(self):
        cmd_line = f"xtb 'constr conf search' gfn2-xtb -f {self.struct('ethanol')} --freeze 1 3"

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))

        objs, outputs = get_input_from_args(args)
        self.assertEqual(len(objs), 1)
        self.assertEqual(len(outputs), 0)

        INPUT_REF = """
        $constrain
        force constant=1.0
        reference=ethanol.xyz
        distance: 1, 3, auto
        atoms: 1,3
        $metadyn
        atoms: 2,4-9"""

        self.assertTrue(self.is_equivalent(INPUT_REF, objs[0].input_file))
        self.assertEqual(
            objs[0].command, "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        )

    @patch("ccinput.utilities.warn")
    def test_warn_unknown(self, warn_fn):
        warn_fn.side_effect = self.get_warn
        line = 'Gaussian sp abcd -bs cc-pvdz --xyz "Cl 0 0 0" -c -1'
        parser = get_parser()
        args = parser.parse_args(shlex.split(line))

        objs, outputs = get_input_from_args(args)

        self.assertNotEqual(len(self.warnings), 0)

    @patch("ccinput.utilities.warn")
    def test_no_warn_utpsstpss(self, warn_fn):
        warn_fn.side_effect = self.get_warn
        line = 'Gaussian sp utpsstpss -bs cc-pvdz --xyz "Cl 0 0 0" -c -1'
        parser = get_parser()
        args = parser.parse_args(shlex.split(line))

        objs, outputs = get_input_from_args(args)

        self.assertEqual(len(self.warnings), 0)

    @patch("ccinput.utilities.warn")
    def test_no_warn_rtpsstpss(self, warn_fn):
        warn_fn.side_effect = self.get_warn
        line = 'Gaussian sp rtpsstpss -bs cc-pvdz --xyz "Cl 0 0 0" -c -1'
        parser = get_parser()
        args = parser.parse_args(shlex.split(line))

        objs, outputs = get_input_from_args(args)

        self.assertEqual(len(self.warnings), 0)

    @patch("ccinput.utilities.warn")
    def test_no_warn_tpsstpss(self, warn_fn):
        warn_fn.side_effect = self.get_warn
        line = 'Gaussian sp tpsstpss -bs cc-pvdz --xyz "Cl 0 0 0" -c -1'
        parser = get_parser()
        args = parser.parse_args(shlex.split(line))

        objs, outputs = get_input_from_args(args)

        self.assertEqual(len(self.warnings), 0)

    def test_synonym_basis_set(self):
        line = 'Gaussian sp tpsstpss -bs ccpvdz --xyz "Cl 0 0 0" -c -1'
        parser = get_parser()
        args = parser.parse_args(shlex.split(line))

        objs, outputs = get_input_from_args(args)


class CliPresetTests(InputTests):
    def test_create_preset(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp tpsstpss -bs ccpvdz --save my_preset"
            cmd(cmd_line=line)

            content = os.listdir(tmp_dir)
            self.assertEqual(len(content), 1)
            self.assertEqual(content[0], "my_preset.preset")

    def test_create_partial_preset(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp -bs ccpvdz --save my_preset"
            cmd(cmd_line=line)

            content = os.listdir(tmp_dir)
            self.assertEqual(len(content), 1)
            self.assertEqual(content[0], "my_preset.preset")

    def test_integrity_preset1(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp tpsstpss -bs ccpvdz --save my_preset"
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["software"], "Gaussian")
            self.assertEqual(preset["type"], "sp")
            self.assertEqual(preset["method"], "tpsstpss")
            self.assertEqual(preset["basis_set"], "ccpvdz")
            self.assertEqual(len(preset.keys()), 5)  # 4 parameters + version

    def test_integrity_preset2(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = 'orca opt m062x -bs def2tzvp --specifications "tightscf" -sm smd -sr smd18 -s methanol --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["software"], "orca")
            self.assertEqual(preset["type"], "opt")
            self.assertEqual(preset["method"], "m062x")
            self.assertEqual(preset["basis_set"], "def2tzvp")
            self.assertEqual(preset["specifications"], "tightscf")
            self.assertEqual(preset["solvation_model"], "smd")
            self.assertEqual(preset["solvation_radii"], "smd18")
            self.assertEqual(preset["solvent"], "methanol")

            self.assertEqual(len(preset.keys()), 9)

    def test_integrity_partial_preset(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = 'orca opt -bs def2tzvp --specifications "tightscf" -sm smd -sr smd18 -s methanol --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["software"], "orca")
            self.assertEqual(preset["type"], "opt")
            self.assertEqual(preset["basis_set"], "def2tzvp")
            self.assertEqual(preset["specifications"], "tightscf")
            self.assertEqual(preset["solvation_model"], "smd")
            self.assertEqual(preset["solvation_radii"], "smd18")
            self.assertEqual(preset["solvent"], "methanol")

            self.assertNotIn("method", preset)

            self.assertEqual(len(preset.keys()), 8)

    def test_structure_not_saved(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = 'orca opt m062x -bs def2tzvp --specifications "tightscf" -sm smd -sr smd18 -s methanol --xyz "Cl 0 0 0" -c -1 --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertNotIn("xyz", preset)
            self.assertEqual(len(preset.keys()), 10)

    def test_override_partially(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = 'orca opt m062x -bs def2tzvp --specifications "tightscf" -sm smd -sr smd18 -s methanol --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["basis_set"], "def2tzvp")
            self.assertEqual(len(preset.keys()), 9)

            line = "orca -bs def2qzvp --save my_preset"
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["software"], "orca")
            self.assertEqual(preset["type"], "opt")
            self.assertEqual(preset["method"], "m062x")
            self.assertEqual(preset["specifications"], "tightscf")
            self.assertEqual(preset["solvation_model"], "smd")
            self.assertEqual(preset["solvation_radii"], "smd18")
            self.assertEqual(preset["solvent"], "methanol")

            self.assertEqual(preset["basis_set"], "def2qzvp")
            self.assertEqual(len(preset.keys()), 9)

    def test_override_completely(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = 'orca opt m062x -bs def2tzvp --specifications "tightscf" -sm smd -sr smd18 -s methanol --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["basis_set"], "def2tzvp")
            self.assertEqual(len(preset.keys()), 9)

            line = 'g16 sp m06 -bs def2qzvp --specifications "opt(maxstep=10) symm" -sm pcm -sr bondi -s acetonitrile --save my_preset'
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertEqual(preset["software"], "g16")
            self.assertEqual(preset["type"], "sp")
            self.assertEqual(preset["method"], "m06")
            self.assertEqual(preset["basis_set"], "def2qzvp")
            self.assertEqual(preset["specifications"], "opt(maxstep=10) symm")
            self.assertEqual(preset["solvation_model"], "pcm")
            self.assertEqual(preset["solvation_radii"], "bondi")
            self.assertEqual(preset["solvent"], "acetonitrile")

            self.assertEqual(len(preset.keys()), 9)

    def test_load_preset(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "orca",
                "type": "opt",
                "method": "m062x",
                "basis_set": "def2tzvp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
            }
            line = "--preset my_preset --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_override1(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "gaussian",
                "type": "opt",
                "method": "m062x",
                "basis_set": "def2tzvp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
            }
            line = "gaussian --preset my_preset --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_override2(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "orca",
                "type": "sp",
                "method": "m06",
                "basis_set": "def2tzvp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
            }
            line = "orca sp m06 --preset my_preset --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_override3(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "orca",
                "type": "opt",
                "method": "m062x",
                "basis_set": "def2svp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
            }
            line = "--preset my_preset --xyz 'Cl 0 0 0' -n 1 --mem 1G -c -1 -bs def2svp"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_override4(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "orca",
                "type": "opt",
                "method": "m062x",
                "basis_set": "def2tzvp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
                "d3": True,
            }
            line = "--xyz 'Cl 0 0 0' -n 1 --preset my_preset --mem 1G -c -1 --d3"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_preset_incomplete(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            line = "--xyz 'Cl 0 0 0' -c -1 -n 1 --mem 1G --preset my_preset"

            with self.assertRaises(SystemExit):
                cmd(cmd_line=line)

    def test_load_preset_complete1(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt -bs def2tzvp --save my_preset"
            cmd(cmd_line=line)

            line = (
                "orca opt m062x --xyz 'Cl 0 0 0' -c -1 -n 1 --mem 1G --preset my_preset"
            )

            cmd(cmd_line=line)

    def test_load_preset_complete2(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x --save my_preset"
            cmd(cmd_line=line)

            args = {
                "software": "orca",
                "type": "opt",
                "method": "m062x",
                "basis_set": "def2tzvp",
                "xyz": "Cl 0 0 0\n",
                "nproc": 1,
                "mem": "1G",
                "charge": -1,
                "d3": True,
            }
            line = "--xyz 'Cl 0 0 0' -n 1 -bs def2tzvp --preset my_preset --mem 1G -c -1 --d3"
            self.assertTrue(self.args_cmd_equivalent(args, line))

    def test_load_unknown_preset(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "orca opt m062x --save my_preset"
            cmd(cmd_line=line)

            line = "--xyz 'Cl 0 0 0' -n 1 -bs def2tzvp --preset other_preset --mem 1G -c -1 --d3"

            with self.assertRaises(SystemExit):
                cmd(cmd_line=line)

    def test_structures_not_saved(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp tpsstpss -bs ccpvdz -f mystruct.xyz --save my_preset"
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertNotIn("file", preset)

    def test_output_not_saved(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp tpsstpss -bs ccpvdz -o calc.com --save my_preset"
            cmd(cmd_line=line)

            with open(os.path.join(tmp_dir, "my_preset.preset")) as f:
                preset = json.load(f)

            self.assertNotIn("output", preset)

    def test_save_and_load(self):
        with tempfile.TemporaryDirectory() as tmp_dir, hide_cmd_output():
            presets.data_dir = tmp_dir
            line = "Gaussian sp tpsstpss -bs ccpvdz -o calc.com --save my_preset --preset my_preset"
            with self.assertRaises(SystemExit):
                cmd(cmd_line=line)



import basis_set_exchange
import numpy as np

from ccinput.utilities import (
    get_method,
    get_solvent,
    get_basis_set,
    clean_xyz,
    get_distance,
    get_angle,
    get_dihedral,
    get_npxyz,
)
from ccinput.constants import CalcType, ATOMIC_NUMBER, LOWERCASE_ATOMIC_SYMBOLS
from ccinput.exceptions import InvalidParameter, ImpossibleCalculation


class GaussianCalculation:
    TEMPLATE = """%chk={}.chk
    %nproc={}
    %mem={}MB
    #p {}

    {}

    {} {}
    {}
    {}
    """

    KEYWORDS = {
        CalcType.OPT: ["opt"],
        CalcType.CONSTR_OPT: ["opt"],
        CalcType.TS: ["opt"],
        CalcType.FREQ: ["freq"],
        CalcType.NMR: ["nmr"],
        CalcType.SP: ["sp"],
        CalcType.UVVIS: ["td"],
        CalcType.UVVIS_TDA: ["tda"],
        CalcType.OPTFREQ: ["opt", "freq"],
    }

    # Get a set of all unique calculation keywords
    KEYWORD_LIST = set([j for i in KEYWORDS.values() for j in i])

    # Number of processors
    # Amount of memory
    # Command line
    # Charge
    # Multiplicity
    # XYZ structure
    # Appendix

    def __init__(self, calc):
        self.calc = calc
        self.has_scan = False
        self.appendix = []
        self.command_line = ""

        self.commands = {}
        self.solvation_radii = {}

        self.confirmed_specifications = ""
        self.xyz_structure = ""
        self.input_file = ""

        if self.calc.type not in self.KEYWORDS:
            raise ImpossibleCalculation(
                f"Gaussian 16 does not support calculations of type {self.calc.type}"
            )

        self.handle_specifications()
        self.handle_command()
        self.handle_xyz()

        self.handle_solvation()

        self.create_input_file()

    def clean(self, s):
        WHITELIST = set(
            "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ/()=-,. "
        )
        return "".join([c for c in s if c in WHITELIST])

    def add_option(self, key, option):
        _option = option.strip()
        if key in self.commands:
            if _option not in self.commands[key]:
                self.commands[key].append(_option)
        else:
            self.commands[key] = [_option]

    def add_options(self, key, options):
        for option in options:
            self.add_option(key, option)

    def add_commands(self, commands):
        for cmd in commands:
            if cmd not in self.commands:
                self.commands[cmd] = []

    def handle_specifications(self):
        s = self.clean(self.calc.parameters.specifications.lower())

        # Could be more sophisticated to catch other incorrect specifications
        if s.count("(") != s.count(")"):
            raise InvalidParameter("Invalid specifications: parenthesis not matching")

        _specifications = ""
        remove = False
        for c in s:
            if c == " " and remove:
                continue
            _specifications += c
            if c == "(":
                remove = True
            elif c == ")":
                remove = False

        for spec in _specifications.split(" "):
            if spec.strip() == "":
                continue
            if spec.find("(") != -1:
                key, options = spec.split("(")
                options = options.replace(")", "")
                for option in options.split(","):
                    if option.strip() != "":
                        self.add_option(key, option)
            elif spec.find("=") != -1:
                try:
                    key, option = spec.split("=")
                except ValueError:
                    raise InvalidParameter(f"Invalid specification: {spec}")
                self.add_option(key, option)
            else:
                self.add_option(spec, "")

        if self.calc.parameters.d3:
            self.add_option("EmpiricalDispersion", "GD3")
        elif self.calc.parameters.d3bj:
            self.add_option("EmpiricalDispersion", "GD3BJ")

    def filter_commands(self, commands):
        """Removes all commands which are not relevant"""
        keys = list(self.commands.keys())
        for key in keys:
            if key in self.KEYWORD_LIST and key not in commands:
                del self.commands[key]

    def handle_command(self):
        self.filter_commands(self.KEYWORDS[self.calc.type])
        self.add_commands(self.KEYWORDS[self.calc.type])

        if self.calc.type == CalcType.TS:
            self.add_options("opt", ["ts", "NoEigenTest", "CalcFC"])
        elif self.calc.type == CalcType.CONSTR_OPT:
            self.add_options("opt", ["modredundant"])

            xyz = get_npxyz(self.calc.xyz)
            gaussian_constraints = ""

            has_scan = False

            if len(self.calc.constraints) == 0:
                raise InvalidParameter("No constraint in constrained optimisation mode")

            for constr in self.calc.constraints:
                if constr.scan:
                    has_scan = True
                gaussian_constraints += constr.to_gaussian()

            self.has_scan = has_scan
            self.appendix.append(gaussian_constraints)

        method = get_method(self.calc.parameters.method, "gaussian")
        basis_set = get_basis_set(self.calc.parameters.basis_set, "gaussian")
        custom_basis_set = self.calc.parameters.custom_basis_sets

        if method == "":
            if self.calc.parameters.theory_level == "hf":
                method = "HF"
            else:
                raise InvalidParameter("No method")

        if basis_set != "":
            if len(custom_basis_set) == 0:
                if self.calc.parameters.density_fitting != "":
                    self.command_line += (
                        f"{method}/{basis_set}/{self.calc.parameters.density_fitting} "
                    )
                else:
                    self.command_line += f"{method}/{basis_set} "
            else:
                gen_keyword, to_append = self.parse_custom_basis_set(basis_set)
                if to_append.strip() == "":
                    self.command_line += f"{method}/{basis_set} "
                else:
                    self.appendix.append(to_append)
                    self.command_line += f"{method}/{gen_keyword} "
        else:
            self.command_line += f"{method} "

    def parse_custom_basis_set(self, base_bs):
        custom_basis_sets = self.calc.parameters.custom_basis_sets
        to_append_gen = []
        to_append_ecp = []

        unique_atoms = []
        normal_atoms = []
        for line in self.calc.xyz.split("\n"):
            if line.strip() == "":
                continue
            a, *_ = line.split()
            if a not in unique_atoms:
                unique_atoms.append(a)
                if a not in normal_atoms and a not in custom_basis_sets:
                    normal_atoms.append(a)

        custom_atoms = []
        ecp = False
        for el, bs_keyword in custom_basis_sets.items():
            if el not in unique_atoms:
                continue

            custom_atoms.append(el)
            try:
                el_num = ATOMIC_NUMBER[el]
            except KeyError:
                raise InvalidParameter(
                    f"Invalid atom in custom basis set string: '{el}'"
                )

            try:
                bs = basis_set_exchange.get_basis(
                    bs_keyword, fmt="gaussian94", elements=[el_num], header=False
                )
            except:
                # Some basis sets are built-in, but use different names as the BSE does (e.g., SDD)
                # In this case, just feed the user keyword in and hope it works.
                # The basis set string has been recognized by ccinput, so it should exist in the program.
                # ECP is added if Z > 18 (Ar)
                to_append_gen.append(f"{el} 0\n{bs_keyword}\n****\n")
                if el_num > 18:
                    ecp = True
                    to_append_ecp.append(f"{el} 0\n{bs_keyword}\n")

            else:
                if bs.find("-ECP") != -1:
                    ecp = True
                    sbs = bs.split("\n")
                    ecp_ind = -1
                    for ind, line in enumerate(sbs):
                        if sbs[ind].find("-ECP") != -1:
                            ecp_ind = ind
                            break
                    bs_gen = "\n".join(sbs[: ecp_ind - 2]) + "\n"
                    bs_ecp = "\n".join(sbs[ecp_ind - 2 :])
                    to_append_gen.append(bs_gen)
                    to_append_ecp.append(bs_ecp)
                else:
                    to_append_gen.append(bs)

        if len(custom_atoms) > 0:
            if ecp:
                gen_keyword = "GenECP"
            else:
                gen_keyword = "Gen"

            custom_bs = ""

            if len(normal_atoms) > 0:
                custom_bs += " ".join(normal_atoms) + " 0\n"
                custom_bs += base_bs + "\n"
                custom_bs += "****\n"

            custom_bs += "".join(to_append_gen) + "\n"
            custom_bs += "".join(to_append_ecp).replace("\n\n", "\n")

            return gen_keyword, custom_bs
        else:
            return self.calc.parameters.basis_set, ""

    def handle_xyz(self):
        lines = [i + "\n" for i in clean_xyz(self.calc.xyz).split("\n") if i != ""]
        self.xyz_structure = "".join(lines)

    def parse_custom_solvation_radii(self):
        for radius in self.calc.parameters.custom_solvation_radii.split(";"):
            if radius.strip() == "":
                continue
            sradius = radius.split("=")
            if len(sradius) != 2:
                raise InvalidParameter(
                    f"Invalid custom solvation radius specification: '{radius}': must follow the pattern '<atom1>=<radius1>;...'"
                )

            element, rad = sradius
            if element not in LOWERCASE_ATOMIC_SYMBOLS:
                raise InvalidParameter(
                    f"Invalid element in custom solvation radius specifications: '{element}'"
                )

            _element = LOWERCASE_ATOMIC_SYMBOLS[element]  # Add the proper case back

            try:
                _rad = float(rad)
            except ValueError:
                raise InvalidParameter(
                    f"Invalid custom solvation radius for element {element}: '{rad}'"
                )
            self.solvation_radii[_element] = _rad

    def get_radii_appendix(self):
        radii_appendix = ""
        for el, rad in self.solvation_radii.items():
            radii_appendix += f"{el} {rad:.2f}\n"
        return radii_appendix

    def handle_solvation(self):
        if self.calc.parameters.solvent.lower() not in ["", "vacuum"]:
            solvent_keyword = get_solvent(
                self.calc.parameters.solvent,
                self.calc.parameters.software,
                solvation_model=self.calc.parameters.solvation_model,
            )

            model = self.calc.parameters.solvation_model
            radii_set = self.calc.parameters.solvation_radii
            custom_radii = self.calc.parameters.custom_solvation_radii

            DEFAULT_RADII_SETS = {
                "smd": ["", "default"],
                "pcm": ["", "uff", "default"],
                "cpcm": ["", "uff", "default"],
            }

            self.add_options("SCRF", [model.upper(), f"Solvent={solvent_keyword}"])

            if radii_set not in DEFAULT_RADII_SETS[model] or custom_radii != "":
                self.add_option("SCRF", "Read")
                self.parse_custom_solvation_radii()

            solv_appendix = ""

            if model == "smd":
                if radii_set == "smd18":
                    # Refined solvation radii
                    # E. Engelage, N. Schulz, F. Heinen, S. M. Huber, D. G. Truhlar,
                    # C. J. Cramer, Chem. Eur. J. 2018, 24, 15983-15987.
                    if "Br" not in self.solvation_radii:
                        self.solvation_radii["Br"] = 2.60
                    if "I" not in self.solvation_radii:
                        self.solvation_radii["I"] = 2.74
            elif model in ["pcm", "cpcm"]:
                if radii_set not in DEFAULT_RADII_SETS[model]:
                    solv_appendix += f"Radii={radii_set}\n"
            else:
                raise InvalidParameter(
                    f"Invalid solvation method for Gaussian: '{self.calc.parameters.solvation_model}'"
                )

            if len(self.solvation_radii) > 0:
                solv_appendix += "modifysph\n\n"
                solv_appendix += self.get_radii_appendix()

            self.appendix.append(solv_appendix)

    def create_input_file(self):
        for cmd, options in self.commands.items():
            option_str = ", ".join(options)
            if option_str != "":
                cmd_formatted = f"{cmd}({option_str}) "
            else:
                cmd_formatted = f"{cmd} "

            if cmd.lower() != "scrf":
                spec_options = []
                for o in options:
                    if o.strip() == "":
                        continue
                    if self.calc.type == CalcType.TS and o in [
                        "ts",
                        "NoEigenTest",
                        "CalcFC",
                    ]:
                        continue
                    elif self.calc.type == CalcType.CONSTR_OPT and o == "modredundant":
                        continue
                    spec_options.append(o)
                if len(spec_options) > 0:
                    spec_option_str = ", ".join(spec_options)

                    confirmed_spec = f"{cmd}({spec_option_str}) "
                    if cmd in self.KEYWORD_LIST:
                        self.confirmed_specifications = (
                            confirmed_spec + self.confirmed_specifications
                        )
                    else:
                        self.confirmed_specifications += confirmed_spec

            # This ensures that the command line follows this pattern:
            # CMD1 <CMD2> METHOD/BASIS_SET <ADDITIONAL_OPTION1> ...
            if cmd in self.KEYWORD_LIST:
                self.command_line = cmd_formatted + self.command_line
            else:
                self.command_line = self.command_line + cmd_formatted

        self.confirmed_specifications = self.confirmed_specifications.strip()

        raw = self.TEMPLATE.format(
            self.calc.name,
            self.calc.nproc,
            self.calc.mem,
            self.command_line.strip(),
            self.calc.header,
            self.calc.charge,
            self.calc.multiplicity,
            self.xyz_structure,
            "\n".join(self.appendix),
        )
        self.input_file = "\n".join([i.strip() for i in raw.split("\n")]).replace(
            "\n\n\n", "\n\n"
        )

    @property
    def output(self):
        return self.input_file



import os
import json
import glob
from pathlib import Path
from appdirs import user_data_dir

from ccinput.__init__ import __version__
from ccinput.calculation import Calculation, Parameters
from ccinput.exceptions import InvalidParameter
from ccinput.utilities import warn

data_dir = user_data_dir("ccinput", "CYLlab")


def warn_invalid_json(preset_name):
    warn(f"Invalid JSON found for existing preset {preset_name} (in {data_dir})")


def save_preset(args, default_args):
    preset_name = args.save

    if len(preset_name.strip()) == 0:
        raise InvalidParameter(
            "The preset name must contain at least one valid character"
        )

    params = vars(args)
    default_params = vars(default_args)

    del params["save"]
    del default_params["save"]

    for k, v in default_params.items():
        if v == params[k] or params[k] is None:
            del params[k]

    if "xyz" in params:
        warn("Ignoring the xyz structure")
        del params["xyz"]

    if "file" in params:
        warn("Ignoring the input structure")
        del params["file"]

    if "output" in params:
        warn("Ignoring the output name")
        del params["output"]

    if "name" in params:
        warn("Ignoring the name")
        del params["name"]

    ## Test the validity of the parameters?

    if not os.path.isdir(data_dir):
        Path(data_dir).mkdir(parents=True, exist_ok=True)

    preset_json = {}

    path = os.path.join(data_dir, f"{preset_name}.preset")
    if os.path.isfile(path):
        try:
            with open(path) as f:
                preset_json = json.load(f)
        except json.decoder.JSONDecodeError:
            warn_invalid_json()

    for k, v in params.items():
        preset_json[k] = v

    preset_json["version"] = __version__

    with open(path, "w") as out:
        json.dump(preset_json, out, indent=4)

    return preset_name


def load_preset(preset_name):
    if len(preset_name.strip()) == 0:
        raise InvalidParameter(
            "The preset name must contain at least one valid character"
        )

    path = os.path.join(data_dir, f"{preset_name}.preset")

    if os.path.isfile(path):
        try:
            with open(path) as f:
                preset_json = json.load(f)
        except json.decoder.JSONDecodeError:
            warn_invalid_json()
    else:
        raise InvalidParameter(
            "No preset found with the name {preset_name} (in {data_dir})"
        )

    return preset_json


def get_preset_names():
    _presets = list(glob.glob(os.path.join(data_dir, "*.preset")))
    return [os.path.splitext(os.path.basename(p))[0] for p in _presets]


def is_preset(preset_name):
    presets = get_preset_names()
    if f"{preset_name}" in presets:
        return True

    print(f"Unknown preset: '{preset_name}'")
    print("")
    return False


def list_presets():
    presets = get_preset_names()
    print("--- Available presets:")
    for p in presets:
        print(p)


def print_preset(preset_name):
    path = os.path.join(data_dir, f"{preset_name}.preset")

    if not os.path.isfile(path):
        raise InvalidParameter(
            f"No preset called {preset_name} found in your user directory ({data_dir})"
        )

    try:
        with open(path) as f:
            preset_json = json.load(f)
    except json.decoder.JSONDecodeError:
        warn_invalid_json(preset_name)
        return

    print(f"--- Saved preset '{preset_name}'")

    for k, v in preset_json.items():
        print(f"{k:<30}{v:<30}")



import os
import string
import numpy as np

from ccinput.constants import *
from ccinput.exceptions import *

MEMORY_FACTORS = {
    "m": 1,
    "mb": 1,
    "mib": 1.048576,  # (1024/1000)^2
    "g": 1000,
    "gb": 1000,
    "gib": 1073.741824,  # 1000*(1024/1000)^3
    "t": 1000000,
    "tb": 1000000,
    "tib": 1099511.628,  # 1000^2*(1024/1000)^4
}


def verify_memory_valid(mem):
    if mem < 0:
        raise InvalidParameter(f"The amount of memory must be positive, not '{mem}'")

    if mem > 1e8:
        raise InvalidParameter(f"Unreasonable amount of memory requested: '{mem}' MB")

    return round(mem)


def standardize_memory(mem):
    """Converts a string specifying an amount of memory to an integer number of megabytes"""

    if isinstance(mem, (int, float)):
        return verify_memory_valid(mem)

    if isinstance(mem, str):
        ind = 0
        for ind, c in enumerate(mem.lower()):
            if c in string.ascii_lowercase:
                break
        else:
            ind += 1

        _val = mem[:ind]
        unit = mem[ind:].lower()

        if len(_val) == 0:
            raise InvalidParameter(f"Invalid memory specification: '{mem}'")

        if len(unit) == 0:
            warn("The amount of memory does not specify a unit; supposing MB")
            unit = "mb"
        else:
            if unit not in MEMORY_FACTORS:
                raise InvalidParameter(
                    f"Unknown unit used for memory specification: '{unit}'"
                )

        try:
            val = float(_val)
        except ValueError:
            raise InvalidParameter(
                f"The amount of memory must be a number (received '{mem}')"
            )
        standard_mem = val * MEMORY_FACTORS[unit]
        return verify_memory_valid(standard_mem)
    else:
        raise InvalidParameter(
            f"Invalid type for the memory specification: {type(mem)}"
        )


def standardize_xyz(xyz):
    """
    Converts variations of the XYZ format into a uniform format for this project.

    Format:
        "El1 X1 Y1 Z2\n El2 X2 Y2 Z2 [...]"

    Example:
        H 0.0 0.0 0.0
        H 1.0 0.0 0.0
    """

    standard_xyz = ""
    if isinstance(xyz, list):
        arr_xyz = xyz
    elif isinstance(xyz, str):
        arr_xyz = xyz.strip().split("\n")
    else:
        raise InvalidParameter(f"Cannot parse xyz from type {type(xyz)}")

    # Check if the xyz contains the two line header
    # Remove it if it does
    try:
        num_atoms = int(arr_xyz[0])
    except ValueError:
        pass
    else:
        if num_atoms == len(arr_xyz) - 2:
            arr_xyz = arr_xyz[2:]
        else:
            raise InvalidXYZ(
                f"Invalid xyz header: {num_atoms} atoms specified, "
                + f"but actually contains {len(arr_xyz) - 2} atoms"
            )

    for el in arr_xyz:
        line_data = []

        if not isinstance(el, str):
            raise InvalidXYZ(
                f"Could not parse xyz from array: contains element '{el}' "
                + "and not string"
            )

        if el.strip() == "":
            continue

        sel = el.strip().split()
        if not len(sel) == 4:
            raise InvalidXYZ(f"Invalid xyz: found line '{el}'")

        if sel[0] not in ATOMIC_NUMBER:
            if sel[0].isdigit():
                try:
                    el_Z = int(sel[0])
                except ValueError:
                    raise InvalidXYZ(f"Invalid atomic label: '{sel[0]}'")
                else:
                    if el_Z not in ATOMIC_SYMBOL:
                        raise InvalidXYZ(f"Invalid atomic number: '{el_Z}'")
                    el_symb = ATOMIC_SYMBOL[el_Z]
                    line_data.append(el_symb)
            else:
                if sel[0].lower() in LOWERCASE_ATOMIC_SYMBOLS:
                    line_data.append(LOWERCASE_ATOMIC_SYMBOLS[sel[0].lower()])
                else:
                    raise InvalidXYZ(f"Invalid atomic label: '{sel[0]}'")
        else:
            line_data.append(sel[0])

        for coord in sel[1:]:
            try:
                c = float(coord)
            except ValueError:
                raise InvalidXYZ(f"Invalid atomic coordinate: '{coord}'")
            else:
                line_data.append(c)

        standard_xyz += "{:<2} {:>12.8f} {:>12.8f} {:>12.8f}\n".format(*line_data)

    return standard_xyz


def parse_xyz_from_file(path):
    if not os.path.isfile(path):
        raise InvalidParameter(f"Input file not found: {path}")

    with open(path) as f:
        lines = f.readlines()

    return standardize_xyz(lines)


def indexify(txt):
    INDEX_CHARS = string.ascii_lowercase + string.digits + "+*"
    _txt = txt.lower()
    return "".join([i for i in _txt if i in INDEX_CHARS])


def get_theory_level(method):
    _method = indexify(method)
    for level, methods in THEORY_LEVELS.items():
        if _method in methods or _method == level:
            return level
    return "dft"


def get_abs_type(str_type):
    """
    Converts a string calculation type into the correct CalcType.
    Takes into account different equivalent ways to write the calculation types.
    """
    _str_type = indexify(str_type)
    for calc_type, synonyms in SYN_TYPES.items():
        if _str_type in synonyms:
            return calc_type

    raise InvalidParameter(f"Invalid calculation type: '{str_type}'")


def get_abs_software(software):
    _software = indexify(software)
    for s in SYN_SOFTWARE:
        if _software in SYN_SOFTWARE[s] or _software == s:
            return s
    raise InvalidParameter(f"Unknown software package: '{software}'")


def get_abs_method(method, trust_me=False):
    _method = indexify(method)
    for m in SYN_METHODS:
        if _method in SYN_METHODS[m] or _method == m:
            return m

    if trust_me:
        warn(f"Using unknown method '{method}'")
        return m
    else:
        raise InvalidParameter(f"Unknown method: '{method}'")


def get_abs_basis_set(basis_set, trust_me=False):
    _bs = indexify(basis_set)

    for bs in SYN_BASIS_SETS:
        if _bs.lower() in SYN_BASIS_SETS[bs] or _bs.lower() == bs:
            return bs
    for bs in BASIS_SET_EXCHANGE_KEY:
        if _bs.lower() == bs:
            return bs
    if trust_me:
        warn(f"Using unknown basis set '{bs}'")
        return bs
    else:
        raise InvalidParameter(f"Unknown basis set: '{basis_set}'")


def get_abs_solvent(solvent, trust_me=False):
    _solvent = indexify(solvent)
    if _solvent in ["", "vacuum", "vac"]:
        return ""
    for solv in SYN_SOLVENTS:
        if _solvent in SYN_SOLVENTS[solv] or _solvent == solv:
            return solv
    if trust_me:
        warn(f"Using unknown solvent '{solvent}'")
        return solvent
    else:
        raise InvalidParameter(f"Unknown solvent: '{solvent}'")


def is_exchange_correlation_combination(method):
    for x in EXCHANGE_FUNCTIONALS:
        if method[: len(x)] == x:
            if method[len(x) :] in CORRELATION_FUNCTIONALS:
                return (
                    EXCHANGE_FUNCTIONALS[method[: len(x)]]
                    + CORRELATION_FUNCTIONALS[method[len(x) :]]
                )
    return False


def get_method(method, software):
    try:
        abs_method = get_abs_method(method)
    except InvalidParameter:
        if software == "gaussian":
            # As far as I know, this kind of specification does not apply to ORCA
            if method.lower()[0] in ["u", "r"]:
                try:
                    abs_method = get_abs_method(method[1:])
                except InvalidParameter:
                    pass
                else:
                    if abs_method in SOFTWARE_METHODS[software]:
                        return (
                            method[0].upper() + SOFTWARE_METHODS[software][abs_method]
                        )

                xc_check = is_exchange_correlation_combination(method.lower()[1:])
                if isinstance(xc_check, str):
                    return method[0].upper() + xc_check

            xc_check = is_exchange_correlation_combination(method.lower())
            if isinstance(xc_check, str):
                return xc_check
        warn(f"Unknown method '{method}'")
        return method
    else:
        if abs_method not in SOFTWARE_METHODS[software]:
            warn(f"Unknown method for this package: '{method}'")
            return method

        return SOFTWARE_METHODS[software][abs_method]


def get_basis_set(basis_set, software):
    try:
        abs_basis_set = get_abs_basis_set(basis_set)
    except InvalidParameter:
        warn(f"Unknown basis set '{basis_set}'")
        return basis_set

    if abs_basis_set in SOFTWARE_BASIS_SETS[software]:
        return SOFTWARE_BASIS_SETS[software][abs_basis_set]
    else:
        return BASIS_SET_EXCHANGE_KEY[abs_basis_set]


def get_solvent(solvent, software, solvation_model="smd"):
    try:
        abs_solvent = get_abs_solvent(solvent)
    except InvalidParameter:
        warn(f"Unknown solvent '{solvent}'")
        return solvent

    if software == "orca" and abs_solvent == "noctanol":
        # Weird exception in ORCA
        if solvation_model == "smd":
            return "1-octanol"
        elif solvation_model == "cpcm":
            return "octanol"
        # Note that ch2cl2 is a valid keyword for SMD, although not listed in the manual

    return SOFTWARE_SOLVENTS[software][abs_solvent]


def has_dispersion_parameters(method, version="d3"):
    try:
        _method = get_abs_method(method)
    except InvalidParameter:
        warn(
            "Unknown method, could not verify if dispersion parameters are available for it"
        )
        return True  # Don't print a second warning

    if _method in FUNCTIONALS_WITH_DISPERSION_PARAMETERS[version]:
        return True
    return False


def clean_xyz(xyz):
    return "".join([x if x in string.printable else " " for x in xyz])


def get_distance(xyz, a, b):
    return np.linalg.norm(xyz[a - 1][1] - xyz[b - 1][1])


def get_angle(xyz, a, b, c):
    v1 = xyz[a - 1][1] - xyz[b - 1][1]
    v2 = xyz[c - 1][1] - xyz[b - 1][1]

    return (
        np.arccos(v1.dot(v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))) * 180 / np.pi
    )


def get_dihedral(xyz, a, b, c, d):
    v1 = xyz[b - 1][1] - xyz[a - 1][1]
    v2 = xyz[c - 1][1] - xyz[b - 1][1]
    v3 = xyz[d - 1][1] - xyz[c - 1][1]

    n1 = np.cross(v1, v2)
    n1 = n1 / np.linalg.norm(n1)

    n2 = np.cross(v2, v3)
    n2 = n2 / np.linalg.norm(n2)

    m1 = np.cross(n1, v2 / np.linalg.norm(v2))
    x = n1.dot(n2)
    y = m1.dot(n2)

    return np.arctan2(y, x) * 180 / np.pi


def get_npxyz(str_xyz):
    lines = [i + "\n" for i in clean_xyz(str_xyz).split("\n")]

    xyz = []
    for line in lines:
        if line.strip() != "":
            a, x, y, z = line.split()
            xyz.append([a, np.array([float(x), float(y), float(z)])])

    return xyz


def get_coord(xyz, ids):
    if len(ids) == 2:
        return get_distance(xyz, *ids)
    elif len(ids) == 3:
        return get_angle(xyz, *ids)
    elif len(ids) == 4:
        return get_dihedral(xyz, *ids)
    else:
        raise InvalidParameter(f"Invalid number of atoms: {len(ids)}")


def warn(msg):
    print(f"*** {msg} ***")


def get_charge_mult_from_name(name):
    charge = 0
    mult = 1

    if "trication" in name:
        charge = 3
    elif "dication" in name:
        charge = 2
    elif "cation" in name:
        charge = 1
    elif "trianion" in name:
        charge = -3
    elif "dianion" in name:
        charge = -2
    elif "anion" in name:
        charge = -1
    elif "neutral" in name:
        charge = 0

    if "radical" in name or "doublet" in name:
        mult = 2
    elif "triplet" in name:
        mult = 3
    elif "singlet" in name:
        mult = 1

    return charge, mult
